package com.persagy.iottransfer.communication.mina.udp.server;

import com.persagy.iottransfer.communication.entity.Packet;
import com.persagy.iottransfer.communication.entity.PacketEntity;
import com.persagy.iottransfer.communication.entity.UDPSendEntity;
import com.persagy.iottransfer.communication.mina.codec.MinaCodecFactory;
import com.persagy.iottransfer.communication.util.*;
import lombok.AllArgsConstructor;
import lombok.Builder;
import lombok.NoArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import org.apache.mina.core.session.IoSession;
import org.apache.mina.filter.codec.ProtocolCodecFactory;
import org.apache.mina.filter.codec.ProtocolCodecFilter;
import org.apache.mina.transport.socket.DatagramSessionConfig;
import org.apache.mina.transport.socket.nio.NioDatagramAcceptor;

import java.io.IOException;
import java.net.InetSocketAddress;
import java.util.Date;
import java.util.HashMap;
import java.util.Map;

@Slf4j
@Builder
@NoArgsConstructor
@AllArgsConstructor
public class UDPServerManager implements IServerManager {
    public final PacketBuffer<PacketEntity> receList = new PacketBuffer<PacketEntity>();
    String ip;
    int port;
    String encoding;
    String aes_password;
    boolean compress = false;
    int Max_size = 1000;
    IServerHandler handlerManager;
    ILogger logger;
    Map<String, Date> connectDateMap = new HashMap<String, Date>();
    boolean allowJustSend = false;

    NioDatagramAcceptor acceptor;
    Map<String, IoSession> socketMap = new HashMap<String, IoSession>();
    int remark_count = 0;

    public UDPServerManager(String ip, int port, String encoding, int Max_size) {
        this.ip = ip;
        this.port = port;
        this.encoding = encoding;
        this.Max_size = Max_size;
    }

    public UDPServerManager(String ip, int port, String encoding, boolean compress, int Max_size) {
        this.ip = ip;
        this.port = port;
        this.encoding = encoding;
        this.compress = compress;
        this.Max_size = Max_size;
    }

    public UDPServerManager(String ip, int port, String encoding, String aes_password, boolean compress, int Max_size) {
        this.ip = ip;
        this.port = port;
        this.encoding = encoding;
        this.aes_password = aes_password;
        this.compress = compress;
        this.Max_size = Max_size;
    }

    @Override
    public void setHandlerManager(IServerHandler handlerManager) {
        this.handlerManager = handlerManager;
    }

    public void setLogger(ILogger logger) {
        this.logger = logger;
    }

    @Override
    public String Name() {
        return "UDP Server " + this.ip + ":" + this.port;
    }

    @Override
    public synchronized void Start() {
        if (this.acceptor == null) {
            try {
                NioDatagramAcceptor acceptorInner = new NioDatagramAcceptor();
                ProtocolCodecFactory factory = new MinaCodecFactory(this.ip, this.port, this.encoding,
                        this.aes_password, this.compress);
                acceptorInner.getFilterChain().addLast("protocol", new ProtocolCodecFilter(factory));
                acceptorInner.setHandler(new UDPServerIoHandler(this));

                DatagramSessionConfig dcfg = acceptorInner.getSessionConfig();
                dcfg.setReuseAddress(true);

                acceptorInner.bind(new InetSocketAddress(this.ip, this.port));

                this.acceptor = acceptorInner;

                log.warn(this.Name() + "\tStart");
            } catch (IOException e) {
                log.error(ExceptionUtil.Instance().GetMessage(e));
            }
        }
    }

    @Override
    public synchronized void Stop() {
        for (int i = 0; i < this.remark_count; i++) {
            this.receList.offer(new PacketEntity(), Integer.MAX_VALUE);
        }
        if (this.acceptor != null) {
            String[] clientAddress_array = this.GetClientAddresss();
            for (String clientAddress : clientAddress_array) {
                try {
                    this.closeClient(clientAddress);
                } catch (Exception e) {
                    e.printStackTrace();
                }
            }
            this.acceptor.unbind();
            this.acceptor.dispose();
            this.acceptor = null;
        }
    }

    @Override
    public synchronized void AllowJustSend_UDP(boolean allow) {
        this.allowJustSend = allow;
    }

    @Override
    public synchronized String[] GetClientAddresss() {
        return this.connectDateMap.keySet().toArray(new String[0]);
    }

    @Override
    public synchronized Date GetConnectTime(String clientAddress) {
        if (this.connectDateMap.containsKey(clientAddress)) {
            return this.connectDateMap.get(clientAddress);
        }
        return null;
    }

    public synchronized void SetConnectTime(String clientAddress, Date date) {
        this.connectDateMap.remove(clientAddress);
        if (date != null) {
            this.connectDateMap.put(clientAddress, date);
        }
    }

    @Override
    public void AppendToSend(String clientAddress, Packet MyPackage) {
        if (this.allowJustSend && !this.socketMap.containsKey(clientAddress)) {
            int index = clientAddress.indexOf(':');
            String remote_ip = clientAddress.substring(1, index);
            int remote_port = Integer.parseInt(clientAddress.substring(index + 1));
            // newSession�ἤ��sessionCreated��sessionOpened
            this.acceptor.newSession(new InetSocketAddress(remote_ip, remote_port),
                    new InetSocketAddress(this.ip, this.port));
        }
        IoSession session = this.socketMap.get(clientAddress);
        if (session != null) {
            session.write(MyPackage);
        }
    }

    public UDPSendEntity PopSend() {
        return null;
    }

    public void AppendToRece(IoSession sa, Packet MyPackage) {
        String clientAddress = sa.getRemoteAddress().toString();
        if (!this.socketMap.containsKey(clientAddress)) {
            this.socketMap.put(clientAddress, sa);
        }
        PacketEntity entity = new PacketEntity();
        entity.address = clientAddress;
        entity.content = MyPackage;
        this.receList.offer(entity, this.Max_size);
    }

    @Override
    public PacketEntity PopRece() {
        PacketEntity entity = this.receList.poll();
        return entity;
    }

    @Override
    public PacketEntity takeRece() throws InterruptedException {
        PacketEntity entity = this.receList.take();
        return entity;
    }

    @Override
    public int ReceBufferSize() {
        return this.receList.BufferSize();
    }

    @Override
    public synchronized void closeClient(String clientAddress) {
        this.socketMap.remove(clientAddress);
    }

    @Override
    public synchronized void hasRemark(int count) {
        remark_count = count;
    }
}
